/// <reference types="node" />
import EventEmitter from 'events';
import { IP2PNode } from '../node';
import { NotaryGroup } from 'tupelo-messages';
import { Transaction } from 'tupelo-messages/transactions/transactions_pb';
import CID from 'cids';
import { IBlockService } from '../chaintree/dag/dag';
import { ICallbackBitswap } from './wrappedbitswap';
import { ChainTree } from '../chaintree';
import Repo from '../repo';
interface IRepo {
    blocks: IBlockService;
}
/**
 * Commmunity is a combination of a Tupelo NotaryGroup and the application 'Community' running.
 * The remote p2p community app will also store blocks, and this is the default supported way
 * that the current wasm-sdk works.
 * @public
 */
export declare class Community extends EventEmitter {
    node: IP2PNode;
    group: NotaryGroup;
    tip?: CID;
    private repo;
    bitswap: ICallbackBitswap;
    blockservice: IBlockService;
    private _started;
    private _startPromise;
    private _startPromiseResolve;
    private _startPromiseReject;
    constructor(node: IP2PNode, group: NotaryGroup, repo: IRepo);
    waitForStart(): Promise<Community>;
    /**
     * getCurrentState returns the current state (signatures)
     * for a given ChainTree (its DID)
     * @public
    */
    getCurrentState(did: string): Promise<import("tupelo-messages/signatures/signatures_pb").TreeState>;
    /**
     * returns the TIP as a CID of the ChainTree. This is more of a convenience function
     * around getting the current state, and then casting the tip, etc.
     * @param did - The DID of the ChainTree
     */
    getTip(did: string): Promise<CID>;
    sendTokenAndGetPayload(tree: ChainTree, tx: Transaction): Promise<import("tupelo-messages").TokenPayload>;
    /**
     * playTransactions is a convenience wrapper on community to make calling the underlying Tupelo.playTransactions
     * easier when using a fully community client
    */
    playTransactions(tree: ChainTree, transactions: Transaction[]): Promise<import("tupelo-messages/signatures/signatures_pb").TreeState>;
    /** next update is a helper function
     * which lets you do an await until the next tip
     * update of the community
    */
    nextUpdate(): Promise<unknown>;
    /**
     * starts up the community
    */
    start(): Promise<Community>;
    stop(): Promise<void>;
    subscribeToTips(): Promise<void>;
}
export declare namespace Community {
    /**
     * getDefault returns the default Tupelo commuity. This is currently the Tupelo testnet.
     * It creates a libp2p node and connects to the testnet, establishes the community connection.
     * This method optionally takes a {@link Repo}. If you do not pass in a Repo, then it will create a
     * default repo. Details on the default can be found:
     * @param repo - (optional) - a {@link Repo} object (wrapper around an IPFS repo).
     * @public
    */
    function getDefault(repo?: Repo): Promise<Community>;
    /**
     * setDefault allows you to set a community you control so that when code calls Community.getDefault() it
     * returns this community. This is useful in situations like local testing, where your test harness can
     * point the code at a local community.
     * @param community - the {@link Community} to set as default
     * @public
     */
    function setDefault(community: Community): void;
    /**
     * fromNotaryGroupToml creates a community from a notary grou TOML string with no boiler plate around
     * creating a new libp2p node and only optionally requires a repo (a new repo will be created if one isn't passed in)
     * @param tomlString - the TOML for the notary group
     * @param repo - (optional) the repo to use for this notary group. Will default to an ondisk repo named after the notary group
     */
    function fromNotaryGroupToml(tomlString: string, repo?: Repo): Promise<Community>;
}
export {};
//# sourceMappingURL=community.d.ts.map