"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = __importDefault(require("events"));
const node_1 = require("../node");
const transactions_pb_1 = require("tupelo-messages/transactions/transactions_pb");
const cids_1 = __importDefault(require("cids"));
const wrappedblockservice_1 = require("./wrappedblockservice");
const tupelo_1 = __importDefault(require("../tupelo"));
const default_1 = require("./default");
const debug_1 = __importDefault(require("debug"));
const repo_1 = __importDefault(require("../repo"));
const notarygroup_1 = __importDefault(require("../notarygroup"));
const debugLog = debug_1.default("community");
// const IpfsBlockService:any = require('ipfs-block-service');
const IpfsBitswap = require('ipfs-bitswap');
const IpfsBlockService = require('ipfs-block-service');
function tipTopicFromNotaryGroup(ng) {
    return ng.getId() + "-tips";
}
/**
 * Commmunity is a combination of a Tupelo NotaryGroup and the application 'Community' running.
 * The remote p2p community app will also store blocks, and this is the default supported way
 * that the current wasm-sdk works.
 * @public
 */
class Community extends events_1.default {
    constructor(node, group, repo) {
        super();
        this._started = false;
        this.node = node;
        this.group = group;
        this.repo = repo;
        this.bitswap = new IpfsBitswap(this.node, this.repo.blocks);
        this.blockservice = new wrappedblockservice_1.WrappedBlockService(new IpfsBlockService(this.repo));
        this.blockservice.setExchange(this.bitswap);
        this._startPromiseResolve = () => { }; // replaced on the line below, this just stops typescript from complaining
        this._startPromiseReject = () => { }; // replaced on the line below, this just stops typescript from complaining
        this._startPromise = new Promise((resolve) => { this._startPromiseResolve = resolve; });
    }
    waitForStart() {
        return __awaiter(this, void 0, void 0, function* () {
            return this._startPromise;
        });
    }
    /**
     * getCurrentState returns the current state (signatures)
     * for a given ChainTree (its DID)
     * @public
    */
    getCurrentState(did) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.start();
            yield this.nextUpdate();
            if (this.tip == undefined) {
                throw new Error("tip still undefined, even though community started and update received");
            }
            return tupelo_1.default.getCurrentState({
                did: did,
                blockService: this.blockservice,
                tip: this.tip,
            });
        });
    }
    /**
     * returns the TIP as a CID of the ChainTree. This is more of a convenience function
     * around getting the current state, and then casting the tip, etc.
     * @param did - The DID of the ChainTree
     */
    getTip(did) {
        return __awaiter(this, void 0, void 0, function* () {
            const state = yield this.getCurrentState(did);
            return new cids_1.default(Buffer.from(state.getNewTip_asU8()));
        });
    }
    sendTokenAndGetPayload(tree, tx) {
        return __awaiter(this, void 0, void 0, function* () {
            const sendTokenPayload = tx.getSendTokenPayload();
            if (tx.getType() != transactions_pb_1.Transaction.Type.SENDTOKEN || sendTokenPayload === undefined) {
                throw new Error("must use a send token transaction here");
            }
            const resp = yield this.playTransactions(tree, [tx]);
            return tupelo_1.default.tokenPayloadForTransaction({
                blockService: this.blockservice,
                tip: tree.tip,
                tokenName: sendTokenPayload.getName(),
                sendId: sendTokenPayload.getId(),
                treeState: resp,
            });
        });
    }
    /**
     * playTransactions is a convenience wrapper on community to make calling the underlying Tupelo.playTransactions
     * easier when using a fully community client
    */
    playTransactions(tree, transactions) {
        return __awaiter(this, void 0, void 0, function* () {
            return tupelo_1.default.playTransactions(this.node.pubsub, this.group, tree, transactions);
        });
    }
    /** next update is a helper function
     * which lets you do an await until the next tip
     * update of the community
    */
    nextUpdate() {
        return __awaiter(this, void 0, void 0, function* () {
            let resolve;
            const p = new Promise((res, _rej) => { resolve = res; });
            this.once('tip', () => { resolve(); });
            return p;
        });
    }
    /**
     * starts up the community
    */
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._started) {
                return this._startPromise;
            }
            this._started = true;
            debugLog("start()");
            this.bitswap.start(() => {
                debugLog("bitswap started");
            });
            if (this.node.isStarted()) {
                try {
                    yield this.subscribeToTips();
                }
                catch (err) {
                    this._started = false;
                    this._startPromiseReject(err);
                }
            }
            else {
                this.node.once('start', () => __awaiter(this, void 0, void 0, function* () {
                    try {
                        yield this.subscribeToTips();
                    }
                    catch (err) {
                        this._started = false;
                        this._startPromiseReject(err);
                    }
                }));
            }
            debugLog("started");
            this._startPromiseResolve(this);
            this.emit('start');
            return this._startPromise;
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            this.bitswap.stop(() => { });
            this.node.stop();
        });
    }
    subscribeToTips() {
        return __awaiter(this, void 0, void 0, function* () {
            let resolve, reject;
            const p = new Promise((res, rej) => { resolve = res, reject = rej; });
            this.node.pubsub.subscribe(tipTopicFromNotaryGroup(this.group), (msg) => {
                if (msg.data.length > 0) {
                    this.tip = new cids_1.default(Buffer.from(msg.data));
                    this.emit('tip', this.tip);
                    debugLog("tip received: cid: ", this.tip, " raw: ", msg.data);
                }
                else {
                    debugLog("received null tip");
                }
            }, (err) => {
                if (err) {
                    reject(err);
                    return;
                }
                debugLog("subscribed to tips");
                resolve();
            });
        });
    }
}
exports.Community = Community;
(function (Community) {
    /**
     * getDefault returns the default Tupelo commuity. This is currently the Tupelo testnet.
     * It creates a libp2p node and connects to the testnet, establishes the community connection.
     * This method optionally takes a {@link Repo}. If you do not pass in a Repo, then it will create a
     * default repo. Details on the default can be found:
     * @param repo - (optional) - a {@link Repo} object (wrapper around an IPFS repo).
     * @public
    */
    function getDefault(repo) {
        return default_1._getDefault(repo);
    }
    Community.getDefault = getDefault;
    /**
     * setDefault allows you to set a community you control so that when code calls Community.getDefault() it
     * returns this community. This is useful in situations like local testing, where your test harness can
     * point the code at a local community.
     * @param community - the {@link Community} to set as default
     * @public
     */
    function setDefault(community) {
        return default_1._setDefault(community);
    }
    Community.setDefault = setDefault;
    /**
     * fromNotaryGroupToml creates a community from a notary grou TOML string with no boiler plate around
     * creating a new libp2p node and only optionally requires a repo (a new repo will be created if one isn't passed in)
     * @param tomlString - the TOML for the notary group
     * @param repo - (optional) the repo to use for this notary group. Will default to an ondisk repo named after the notary group
     */
    function fromNotaryGroupToml(tomlString, repo) {
        return new Promise((res, rej) => __awaiter(this, void 0, void 0, function* () {
            const ng = notarygroup_1.default(tomlString);
            try {
                const node = yield node_1.p2p.createNode({ bootstrapAddresses: ng.getBootstrapAddressesList() });
                if (repo == undefined) {
                    repo = new repo_1.default(ng.getId());
                    try {
                        yield repo.init({});
                        yield repo.open();
                    }
                    catch (e) {
                        rej(e);
                    }
                }
                const c = new Community(node, ng, repo.repo);
                node.start(() => __awaiter(this, void 0, void 0, function* () {
                    res(yield c.start());
                }));
            }
            catch (e) {
                debugLog("error creating community: ", e);
                rej(e);
            }
        }));
    }
    Community.fromNotaryGroupToml = fromNotaryGroupToml;
})(Community = exports.Community || (exports.Community = {}));
