"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const tupelo_1 = __importDefault(require("../tupelo"));
const community_pb_1 = require("tupelo-messages/community/community_pb");
const debug_1 = __importDefault(require("debug"));
const any_pb_js_1 = require("google-protobuf/google/protobuf/any_pb.js");
const log = debug_1.default("community:messaging");
/**
 * CommunityMessenger handles the overlay pubsub community provides to make sure
 * that nodes correctly relay messages within this community.
 * Community listends to a number of libp2p topics (we call shards here) and internal community topics
 * are broadcast on those libp2p topics so that libp2p will correctly relay those messages.
 * @beta
 */
class CommunityMessenger {
    constructor(name, shards, key, localIdentifier, pubsub) {
        this.name = name;
        this.shards = shards;
        this.pubsub = pubsub;
        this.localIdentifier = localIdentifier;
        this.key = key;
        this.subscriptions = new Map();
    }
    /**
     *
     * @param topic - the topic to broadcast to (the community topic, not the pubsub topic)
     * @param payload - the data to send
     * @beta
     */
    publish(topic, payload) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                log("publish called for ", topic);
                const communityTopic = yield this._communityTopic(topic);
                const env = new community_pb_1.Envelope();
                env.setFrom(this.localIdentifier);
                env.setPayload(payload);
                env.setTopicsList([Buffer.from(topic)]);
                env.setTo(Buffer.from(communityTopic));
                try {
                    const bits = yield tupelo_1.default.getSendableEnvelopeBytes(env, this.key);
                    this.pubsub.publish(communityTopic, bits, (err) => {
                        if (err === undefined) {
                            log("error publishing: ", err);
                            reject(err);
                            return;
                        }
                        log("published message to: ", topic, " community topic: ", communityTopic);
                        resolve();
                    });
                }
                catch (e) {
                    log("rejecting because: ", e);
                    reject(e);
                }
            }));
        });
    }
    /**
     *
     * @param topic - the community topic to subscribe to
     * @param subscriber - a callback function which will be called with a protobuf instance of Envelope ( the js protobuf object, not the bytes )
     * @beta
     */
    subscribe(topic, subscriber) {
        return __awaiter(this, void 0, void 0, function* () {
            let resolve, reject;
            const p = new Promise((res, rej) => { resolve = res; reject = rej; });
            const communityTopic = yield this._communityTopic(topic);
            log("subscribing to: ", communityTopic);
            const subFunc = (msg) => {
                const any = any_pb_js_1.Any.deserializeBinary(msg.data);
                const env = community_pb_1.Envelope.deserializeBinary(any.getValue_asU8());
                if (topicsListIncludes(env.getTopicsList_asU8(), topic)) {
                    log("message matches requested topic");
                    subscriber(env);
                }
            };
            let existingSubs = this.subscriptions.get(topic);
            if (existingSubs === undefined) {
                existingSubs = [];
            }
            existingSubs.push({ originalFn: subscriber, subFn: subFunc });
            this.subscriptions.set(topic, existingSubs);
            this.pubsub.subscribe(communityTopic, subFunc, (err) => {
                if (err !== undefined) {
                    log("error subscribing: ", err);
                    reject(err);
                    return;
                }
                log("subscribed to: ", communityTopic);
                resolve();
            });
            return p;
        });
    }
    /**
     * unsubscribe is the reverse of subscribe and works the same as libp2p pubsub works,
     * you pass in a topic and the original subscription and your callback will no longer
     * receive messages from that topic
     * @param topic - the community topic used to subscribe
     * @param subscriber - the callback function your originally passed to subscribe
     * @beta
     */
    unsubscribe(topic, subscriber) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                let subscriptions = this.subscriptions.get(topic);
                if (subscriptions === undefined) {
                    resolve();
                    return;
                }
                let communityTopic;
                try {
                    communityTopic = yield this._communityTopic(topic);
                }
                catch (err) {
                    reject(err);
                    return;
                }
                const index = this.indexOfSubscriber(subscriptions, subscriber);
                if (index === -1) {
                    resolve();
                    return;
                }
                const subHolder = subscriptions[index];
                this.subscriptions.set(topic, subscriptions.splice(index, 1));
                this.pubsub.unsubscribe(communityTopic, subHolder.subFn, resolve);
            }));
        });
    }
    _communityTopic(topic) {
        return __awaiter(this, void 0, void 0, function* () {
            const shardNumber = yield tupelo_1.default.hashToShardNumber(topic, this.shards);
            return this.name + "-" + shardNumber;
        });
    }
    indexOfSubscriber(subscriptions, subscriber) {
        var i = subscriptions.length;
        while (i--) {
            if (subscriptions[i].originalFn === subscriber) {
                return i;
            }
        }
        return -1;
    }
}
exports.CommunityMessenger = CommunityMessenger;
const topicsListIncludes = (list, topic) => {
    for (let i of list) {
        if (Buffer.from(i).toString() == topic) {
            return true;
        }
    }
    return false;
};
exports.default = CommunityMessenger;
