import CID from 'cids';
import { Transaction, Envelope } from 'tupelo-messages';
import { TokenPayload } from 'tupelo-messages/transactions/transactions_pb';
import { IBlockService } from './chaintree/dag/dag';
import ChainTree from './chaintree/chaintree';
import { TreeState, Signature } from 'tupelo-messages/signatures/signatures_pb';
import { NotaryGroup } from 'tupelo-messages/config/config_pb';
import { EcdsaKey } from './crypto';
/**
 * The interface describing libp2p pubsub
 * @public
 */
export interface IPubSub {
    publish(topic: string, data: Uint8Array, cb: Function): null;
    subscribe(topic: string, onMsg: Function, cb: Function): null;
    unsubscribe(topic: string, onMsg: Function, cb: Function): null;
}
interface IGetCurrentStateOptions {
    blockService: IBlockService;
    tip: CID;
    did: string;
}
interface ITransactionPayloadOpts {
    blockService: IBlockService;
    tip: CID;
    tokenName: string;
    sendId: string;
    treeState: TreeState;
}
/**
 * Tupelo is the more "raw" namespace, it is generally expected
 * that you would use higher level wrapper classes around this namespace.
 * For example: {@link EcdsaKey} or {@link Community#"class"}
 * @public
 */
export declare namespace Tupelo {
    function generateKey(): Promise<Uint8Array[]>;
    function passPhraseKey(phrase: Uint8Array, salt: Uint8Array): Promise<Uint8Array[]>;
    function keyFromPrivateBytes(bytes: Uint8Array): Promise<Uint8Array[]>;
    function ecdsaPubkeyToDid(pubKey: Uint8Array): Promise<string>;
    function ecdsaPubkeyToAddress(pubKey: Uint8Array): Promise<string>;
    function getCurrentState(opts: IGetCurrentStateOptions): Promise<TreeState>;
    function newEmptyTree(store: IBlockService, publicKey: Uint8Array): Promise<CID>;
    /**
     * signMessage allows an EcdsaKey to sign an arbitrary message
     * @param key - the EcdsaKey to sign the message (must have a privateKey)
     * @param message - the message to sign (arbitrary Uint8Arry/Buffer of bytes)
     * @public
     */
    function signMessage(key: EcdsaKey, message: Uint8Array): Promise<Signature>;
    /**
     * verifyMessage is the recipricol of {@link signMessage}, it verifies that a signature is valid.
     * @param address - the string address to verify against ( see {@link ecdsaPubkeyToAddress} )
     * @param message - the message to verify ( see {@link signMessage} )
     * @param signature - the {@link Signature} object to verify
     * @public
     */
    function verifyMessage(address: string, message: Uint8Array, signature: Signature): Promise<boolean>;
    function tokenPayloadForTransaction(opts: ITransactionPayloadOpts): Promise<TokenPayload>;
    function hashToShardNumber(topic: string, maxShards: number): Promise<number>;
    function getSendableEnvelopeBytes(env: Envelope, key: EcdsaKey): Promise<Uint8Array>;
    /**
     * verifyCurrentState takes a notary group and a currentstate and makes sure it meets the requirements
     * for being a valid signed currentState (2/3 of signers, valid signature, etc)
     * @param notaryGroup - the protobuf version of the notary group config
     * @param state - the CurrentState (often returned by a playTransactions)
     * @public
     */
    function verifyCurrentState(notaryGroup: NotaryGroup, state: TreeState): Promise<boolean>;
    function playTransactions(publisher: IPubSub, notaryGroup: NotaryGroup, tree: ChainTree, transactions: Transaction[]): Promise<TreeState>;
}
export default Tupelo;
//# sourceMappingURL=tupelo.d.ts.map