package gossip

import (
	"testing"

	"github.com/ethereum/go-ethereum/common/hexutil"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ipfs/go-cid"
	"github.com/quorumcontrol/chaintree/safewrap"
	"github.com/quorumcontrol/messages/v2/build/go/services"
	"github.com/quorumcontrol/tupelo/sdk/gossip/testhelpers"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func newPopulatedMempool() (*mempool, cid.Cid, *services.AddBlockRequest) {
	sw := &safewrap.SafeWrap{}
	abr := &services.AddBlockRequest{
		ObjectId: []byte("test"),
	}
	n := sw.WrapObject(abr)

	pool := newMempool()
	pool.Add(&AddBlockWrapper{AddBlockRequest: abr, cid: n.Cid()})
	return pool, n.Cid(), abr
}

func TestMempoolContains(t *testing.T) {
	sw := &safewrap.SafeWrap{}

	pool, id, _ := newPopulatedMempool()
	badID := sw.WrapObject("not there").Cid()

	assert.True(t, pool.Contains(id))
	assert.False(t, pool.Contains(badID))
	assert.False(t, pool.Contains(id, badID))
}

func TestMempoolGet(t *testing.T) {
	sw := &safewrap.SafeWrap{}

	pool, id, _ := newPopulatedMempool()
	badID := sw.WrapObject("not there").Cid()

	assert.NotNil(t, pool.Get(id))
	assert.Nil(t, pool.Get(badID))
}

func TestMempoolDeleteIDAndConflictSet(t *testing.T) {
	t.Run("it deletes a single id", func(t *testing.T) {
		pool, id, _ := newPopulatedMempool()

		assert.NotNil(t, pool.Get(id))
		pool.DeleteIDAndConflictSet(id)
		assert.Nil(t, pool.Get(id))
		assert.False(t, pool.Contains(id))
		assert.Equal(t, 0, pool.Length())
	})

	t.Run("it deletes all ids from conflictset as well", func(t *testing.T) {
		treeKey, err := crypto.GenerateKey()
		require.Nil(t, err)

		sw := &safewrap.SafeWrap{}
		abr1 := testhelpers.NewValidTransactionWithPathAndValue(t, treeKey, "/path", "value")
		abr2 := testhelpers.NewValidTransactionWithPathAndValue(t, treeKey, "/path", "differentvalue")
		abr1Cid := sw.WrapObject(abr1).Cid()
		abr2Cid := sw.WrapObject(abr2).Cid()
		t.Logf("abr1: %s, abr2: %s", abr1Cid.String(), abr2Cid.String())

		// abr1 and abr2 are conflicting blocks added to the same chaintree
		require.Equal(t, abr1.ObjectId, abr2.ObjectId)
		require.Equal(t, abr1.Height, abr2.Height)
		require.NotEqual(t, abr1.NewTip, abr2.NewTip)

		pool := newMempool()
		pool.Add(&AddBlockWrapper{AddBlockRequest: &abr1, cid: abr1Cid})
		pool.Add(&AddBlockWrapper{AddBlockRequest: &abr2, cid: abr2Cid})
		assert.True(t, pool.Contains(abr1Cid, abr2Cid))
		pool.DeleteIDAndConflictSet(abr2Cid)
		assert.False(t, pool.Contains(abr1Cid))
		assert.False(t, pool.Contains(abr2Cid))
	})
}

func TestMempoolPreferred(t *testing.T) {
	sw := &safewrap.SafeWrap{}

	// use a canonical key because that influences the hash of the ABRs
	keyHex := "0xcd24af4d6c47530202f00442282fa23e06c1adea93e0264cacabf274241918d2"
	treeKey, err := crypto.ToECDSA(hexutil.MustDecode(keyHex))
	require.Nil(t, err)

	abr1 := testhelpers.NewValidTransactionWithPathAndValue(t, treeKey, "/path", "value")
	abr2 := testhelpers.NewValidTransactionWithPathAndValue(t, treeKey, "/path", "differentvalue")

	abr1Cid := sw.WrapObject(abr1).Cid()
	abr2Cid := sw.WrapObject(abr2).Cid()
	t.Logf("abr1: %s, abr2: %s", abr1Cid.String(), abr2Cid.String())

	// the preferred transaction is the one with the lower CID
	winningCID := abr1Cid
	losingCID := abr2Cid
	if abr2Cid.String() <= abr1Cid.String() {
		winningCID = abr2Cid
		losingCID = abr1Cid
	}

	// abr1 and abr2 are conflicting blocks added to the same chaintree
	require.Equal(t, abr1.ObjectId, abr2.ObjectId)
	require.Equal(t, abr1.Height, abr2.Height)
	require.NotEqual(t, abr1.NewTip, abr2.NewTip)

	pool := newMempool()
	pool.Add(&AddBlockWrapper{AddBlockRequest: &abr1, cid: abr1Cid})
	require.True(t, preferredContains(pool, abr1Cid))

	pool.Add(&AddBlockWrapper{AddBlockRequest: &abr2, cid: abr2Cid})
	assert.True(t, preferredContains(pool, winningCID))
	assert.False(t, preferredContains(pool, losingCID))
}

func preferredContains(pool *mempool, id cid.Cid) bool {
	for _, pID := range pool.Preferred() {
		if pID.Equals(id) {
			return true
		}
	}
	return false
}
